import numpy as np

from hstrat._auxiliary_lib import cmp_approx


def test_cmp_approx_returns_0_for_close_values():
    assert cmp_approx(1.0, 1.0000000001) == 0
    assert cmp_approx(1.0000000001, 1.0) == 0


def test_cmp_approx_returns_1_for_a_greater_than_b():
    assert cmp_approx(2.0, 1.0) == 1


def test_cmp_approx_returns_minus_1_for_a_less_than_b():
    assert cmp_approx(1.0, 2.0) == -1


def test_cmp_approx_returns_0_for_equal_values():
    assert cmp_approx(2.0, 2.0) == 0


def test_cmp_approx_returns_0_for_close_values_numpy():
    assert cmp_approx(np.double(1.0), np.double(1.0000000001)) == 0
    assert cmp_approx(np.double(1.0000000001), np.double(1.0)) == 0


def test_cmp_approx_returns_1_for_a_greater_than_b_numpy():
    assert cmp_approx(np.double(2.0), np.double(1.0)) == 1


def test_cmp_approx_returns_minus_1_for_a_less_than_b_numpy():
    assert cmp_approx(np.double(1.0), np.double(2.0)) == -1


def test_cmp_approx_returns_0_for_equal_values_numpy():
    assert cmp_approx(np.double(2.0), np.double(2.0)) == 0



import numbers

import pandas as pd

from ._alifestd_mark_origin_time_delta_asexual import (
    alifestd_mark_origin_time_delta_asexual,
)


def alifestd_sum_origin_time_deltas_asexual(
    phylogeny_df: pd.DataFrame,
    mutate: bool = False,
) -> numbers.Number:
    """Sum differences between taxa origin times and their ancestors' origin
    time.

    Input dataframe is not mutated by this operation unless `mutate` set True.
    """

    if "origin_time_delta" not in phylogeny_df.columns:
        if not mutate:
            phylogeny_df = phylogeny_df.copy()

        phylogeny_df = alifestd_mark_origin_time_delta_asexual(
            phylogeny_df, mutate=True
        )

    return phylogeny_df["origin_time_delta"].sum()



import typing

caretup_marker: typing.List[typing.Tuple[int, int]] = [
    (0, 0),
    (-1, -(3**0.5)),
    (1, -(3**0.5)),
    (0, 0),
]
"""A equilateral triangle, oriented tip up and positioned within the
encapsulating unit cell so that it appears just below the coordinates it is
placed at.

Equivalent to matplotlib "CARETUP" marker, except that it is a closed path so
edge outline will appear properly.
"""



import typing


# Adapted from https://stackoverflow.com/a/3787989
def all_same(items: typing.Iterable) -> bool:
    """Do all items in items compare equal?

    Returns True if items is empty.
    """
    it = iter(items)

    for first in it:
        break
    else:
        return True  # empty case, note all([]) is True

    return all(x == first for x in it)



import operator
import typing

from ..._auxiliary_lib import HereditaryStratigraphicArtifact
from ._calc_rank_of_earliest_detectable_mrca_between import (
    calc_rank_of_earliest_detectable_mrca_between,
)
from ._calc_rank_of_mrca_bounds_between import calc_rank_of_mrca_bounds_between


def calc_rank_of_mrca_uncertainty_between(
    first: HereditaryStratigraphicArtifact,
    second: HereditaryStratigraphicArtifact,
    prior: typing.Literal["arbitrary"],
    confidence_level: float = 0.95,
) -> typing.Optional[int]:
    """How wide is the estimate window for generation of MRCA?

    Calculate uncertainty of estimate for the number of depositions
    elapsed along the line of descent before the most common recent
    ancestor with second.

    Returns 0 if no common ancestor between first and second can be resolved
    with sufficient confidence. If insufficient common ranks between first
    and second are available to resolve any common ancestor, returns None.

    See Also
    --------
    calc_rank_of_mrca_bounds_between :
        Calculates bound whose uncertainty this method reports. See the
        corresponding docstring for explanation of parameters.
    """
    if (
        calc_rank_of_earliest_detectable_mrca_between(
            first,
            second,
            confidence_level=confidence_level,
        )
        is None
    ):
        return None

    bounds = calc_rank_of_mrca_bounds_between(
        first,
        second,
        prior=prior,
        confidence_level=confidence_level,
    )
    # must coerce to int because
    # numpy unsigned types wraparound when subtracting larger from smaller
    return 0 if bounds is None else abs(operator.sub(*map(int, bounds))) - 1



"""Frozen representations of genome annotations for efficient postprocessing
and analysis."""
from .._auxiliary_lib import lazy_attach_stub

__getattr__, __dir__, __all__ = lazy_attach_stub(__name__, __file__)
del lazy_attach_stub



from numbers import Number
import typing


def demark(object_: typing.Any) -> int:
    """Provide a hashable identifier for `object_`."""
    return object_ if isinstance(object_, (str, Number)) else id(object_)



import typing

import anytree


class AnyTreeFastPreOrderIter:
    r"""Iterator that traverses an `anytree.AnyNode` tree using pre-order
    strategy (self, children).

    Parameters
    ----------
    root : anytree.AnyNode
        The root node of the tree to iterate over.

    Examples
    --------
    Given the following tree:

        a
       / \
      b   c
     / \   \
    d   e   f

    We can create an instance of `AnyTreeFastPreOrderIter` and use it to
    iterate over the tree:

    >>> from anytree import Node
    >>> root = Node("a")
    >>> b = Node("b", parent=root)
    >>> c = Node("c", parent=root)
    >>> d = Node("d", parent=b)
    >>> e = Node("e", parent=b)
    >>> f = Node("f", parent=c)
    >>> for node in AnyTreeFastPreOrderIter(root):
    ...     print(node.name)
    ...
    a
    b
    c
    d
    e
    f

    Notes
    -----
    This implementation is faster than the `LevelOrderIter` implementation
    provided by `anytree` (especially for large trees). Because it is iterative
    instead of recursive, it won't cause recursion limit errors.
    """

    _stack: typing.List[anytree.AnyNode]

    def __init__(
        self: "AnyTreeFastPreOrderIter", root: anytree.AnyNode
    ) -> None:
        """Initialize the iterator.

        Parameters
        ----------
        root : anytree.AnyNode
            The root node of the tree to iterate over.
        """
        self._stack = [root]

    def __iter__(self: "AnyTreeFastPreOrderIter") -> "AnyTreeFastPreOrderIter":
        """Return the iterator object itself.

        Returns
        -------
        AnyTreeFastPreOrderIter
            The iterator object itself.
        """
        return self

    def __next__(self: "AnyTreeFastPreOrderIter") -> anytree.AnyNode:
        """Return the next node in pre-order traversal.

        Returns
        -------
        anytree.AnyNode
            The next node in pre-order traversal.

        Raises
        ------
        StopIteration
            When there are no more nodes to traverse.
        """
        if not self._stack:
            raise StopIteration

        node = self._stack.pop()

        if node.children:
            self._stack.extend(reversed(node.children))

        return node



import pandas as pd

from ._alifestd_convert_root_ancestor_token import (
    alifestd_convert_root_ancestor_token,
)


def alifestd_assign_root_ancestor_token(
    phylogeny_df: pd.DataFrame,
    root_ancestor_token: str,
    mutate: bool = False,
) -> pd.DataFrame:
    """Set `root_ancestor_token` for "ancestor_list" column.

    The option `root_ancestor_token` will be sandwiched in brackets to create
    the ancestor list entry for genesis organisms. For example, the token
    "None" will yield the entry "[None]" and the token "" will yield the entry
    """

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    phylogeny_df["ancestor_list"] = alifestd_convert_root_ancestor_token(
        phylogeny_df["ancestor_list"],
        root_ancestor_token,
        mutate=False,  # prevent assign to slice warning
    )

    return phylogeny_df



from collections import Counter
import os

import alifedata_phyloinformatics_convert as apc
import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_aggregate_phylogenies,
    alifestd_find_leaf_ids,
    alifestd_has_contiguous_ids,
    alifestd_is_asexual,
    alifestd_is_topologically_sorted,
    alifestd_parse_ancestor_ids,
    alifestd_to_working_format,
    alifestd_validate,
)

assets_path = os.path.join(os.path.dirname(__file__), "assets")


@pytest.mark.parametrize(
    "phylogeny_df",
    [
        pd.read_csv(
            f"{assets_path}/example-standard-toy-sexual-phylogeny.csv"
        ),
        alifestd_aggregate_phylogenies(
            [
                pd.read_csv(
                    f"{assets_path}/example-standard-toy-sexual-phylogeny.csv"
                ),
                pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
            ]
        ),
        pd.read_csv(
            f"{assets_path}/example-standard-toy-asexual-phylogeny.csv"
        ),
        pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
        pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
        alifestd_aggregate_phylogenies(
            [
                pd.read_csv(f"{assets_path}/nk_ecoeaselection.csv"),
                pd.read_csv(f"{assets_path}/nk_lexicaseselection.csv"),
            ]
        ),
        pd.read_csv(f"{assets_path}/nk_tournamentselection.csv"),
    ],
)
def test_alifestd_to_working_format(phylogeny_df):
    phylogeny_df_ = phylogeny_df.copy()
    working_df = alifestd_to_working_format(phylogeny_df)
    assert alifestd_validate(working_df)
    # check for side effects
    assert phylogeny_df.equals(phylogeny_df_)

    assert alifestd_has_contiguous_ids(working_df)
    assert alifestd_is_topologically_sorted(working_df)

    assert len(phylogeny_df) == len(working_df)
    assert len(alifestd_find_leaf_ids(phylogeny_df)) == len(
        alifestd_find_leaf_ids(working_df)
    )

    if alifestd_is_asexual(phylogeny_df):
        phylogeny_trees = apc.alife_dataframe_to_dendropy_trees(phylogeny_df)
        working_trees = apc.alife_dataframe_to_dendropy_trees(working_df)

        assert Counter(
            node.level() for tree in phylogeny_trees for node in tree
        ) == Counter(node.level() for tree in working_trees for node in tree)
        assert Counter(
            len(node.child_nodes())
            for tree in phylogeny_trees
            for node in tree
        ) == Counter(
            len(node.child_nodes()) for tree in working_trees for node in tree
        )
        assert Counter(
            (node.level(), len(node.child_nodes()))
            for tree in phylogeny_trees
            for node in tree
        ) == Counter(
            (node.level(), len(node.child_nodes()))
            for tree in working_trees
            for node in tree
        )

        assert "ancestor_id" in working_df
        assert set(
            working_df.loc[
                working_df["ancestor_id"] != working_df["id"],
                "ancestor_id",
            ]
        ) == set(
            ancestor_id
            for ancestor_list_str in working_df["ancestor_list"]
            for ancestor_id in alifestd_parse_ancestor_ids(ancestor_list_str)
        )



import matplotlib as mpl


# adapted from https://stackoverflow.com/a/42156450
class ScalarFormatterFixedPrecision(mpl.ticker.ScalarFormatter):

    _precision: int

    def __init__(
        self: "ScalarFormatterFixedPrecision",
        precision: int = 1,
    ) -> None:
        super(ScalarFormatterFixedPrecision, self).__init__()
        self._precision = precision

    # Override function that finds format to use.
    def _set_format(
        self: "ScalarFormatterFixedPrecision",
    ) -> None:
        self.format = "%.1f"



import networkx as nx

import hstrat.phylogenetic_inference.tree._impl as impl


def test_estimate_origin_times_with_simple_tree():
    graph = nx.Graph()
    graph.add_edge("A", "B", length=8)
    graph.add_edge("B", "C", length=2)
    graph.add_edge("C", "D", length=1)
    graph.add_edge("C", "E", length=11)
    leaf_node_origin_times = {"D": 52, "E": 62}

    for node in graph.nodes:
        graph.nodes[node]["taxon_label"] = node

    node_origin_times = impl.estimate_origin_times(
        graph, leaf_node_origin_times
    )
    assert node_origin_times == {
        "E": 62,
        "D": 52,
        "C": 51,
        "B": 49,
        "A": 41,
    }


def test_estimate_origin_times_with_inconsistent_tree():
    graph = nx.Graph()
    graph.add_edge("A", "B", length=8)
    graph.add_edge("B", "C", length=5)
    graph.add_edge("C", "D", length=1)
    graph.add_edge("C", "E", length=10)
    leaf_node_origin_times = {"D": 54, "E": 60}

    for node in graph.nodes:
        graph.nodes[node]["taxon_label"] = node

    node_origin_times = impl.estimate_origin_times(
        graph, leaf_node_origin_times
    )
    for from_, to in graph.edges:
        assert 0 <= node_origin_times[from_] <= node_origin_times[to] <= 60


def test_estimate_origin_times_with_empty_tree():
    graph = nx.Graph()
    leaf_node_origin_times = {}
    node_origin_times = impl.estimate_origin_times(
        graph, leaf_node_origin_times
    )
    assert node_origin_times == {}


def test_estimate_origin_times_with_singleton():
    graph = nx.Graph()
    graph.add_node("Q")
    leaf_node_origin_times = {"Q": 0}
    for node in graph.nodes:
        graph.nodes[node]["taxon_label"] = node

    node_origin_times = impl.estimate_origin_times(
        graph, leaf_node_origin_times
    )
    assert node_origin_times == {"Q": 0}


def test_estimate_origin_times_with_disjoint_tree1():
    graph = nx.Graph()
    graph.add_edge("A", "B", length=8)
    graph.add_edge("B", "C", length=2)
    graph.add_edge("C", "D", length=1)
    graph.add_edge("C", "E", length=11)
    graph.add_node("Q")
    leaf_node_origin_times = {"D": 52, "E": 62, "Q": 17}

    for node in graph.nodes:
        graph.nodes[node]["taxon_label"] = node

    node_origin_times = impl.estimate_origin_times(
        graph, leaf_node_origin_times
    )
    assert node_origin_times == {
        "A": 41,
        "B": 49,
        "C": 51,
        "D": 52,
        "E": 62,
        "Q": 17,
    }


def test_estimate_origin_times_with_disjoint_tree2():
    graph = nx.Graph()
    graph.add_edge("A", "B", length=8)
    graph.add_edge("B", "C", length=2)
    graph.add_edge("C", "D", length=1)
    graph.add_edge("C", "E", length=11)
    graph.add_edge("Q", "R", length=12)
    leaf_node_origin_times = {"D": 52, "E": 62, "Q": 17}

    for node in graph.nodes:
        graph.nodes[node]["taxon_label"] = node

    node_origin_times = impl.estimate_origin_times(
        graph, leaf_node_origin_times
    )
    assert node_origin_times == {
        "A": 41,
        "B": 49,
        "C": 51,
        "D": 52,
        "E": 62,
        "R": 5,
        "Q": 17,
    }
